/*
 * Copyright 2010-2015 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.jetbrains.kotlin.incremental.storage

import org.jetbrains.annotations.TestOnly
import java.io.Closeable
import java.io.File
import java.io.IOException

open class BasicMapsOwner(val cachesDir: File) : Closeable {
    private val maps = arrayListOf<BasicMap<*, *>>()

    companion object {
        val CACHE_EXTENSION = "tab"
    }

    protected val String.storageFile: File
        get() = File(cachesDir, this + "." + CACHE_EXTENSION)

    @Synchronized
    protected fun <K, V, M : BasicMap<K, V>> registerMap(map: M): M {
        maps.add(map)
        return map
    }

    fun flush() {
        forEachMapSafe("flush", BasicMap<*, *>::flush)
    }

    override fun close() {
        forEachMapSafe("close", BasicMap<*, *>::close)
    }

    open fun deleteStorageFiles() {
        forEachMapSafe("deleteStorageFiles", BasicMap<*, *>::deleteStorageFiles)
    }

    /**
     * DEPRECATED: This API should be removed because
     *   - It's not clear what [memoryCachesOnly] means.
     *   - In the past, when `memoryCachesOnly=true` we applied a small optimization: Checking
     *   [com.intellij.util.io.PersistentHashMap.isDirty] before calling [com.intellij.util.io.PersistentHashMap.force]. However, if that
     *   optimization is useful, it's better to always do it (perhaps inside the [com.intellij.util.io.PersistentHashMap.force] method
     *   itself) rather than doing it based on the value of this parameter.
     *
     * Instead, just call [flush] (without a parameter) directly.
     */
    fun flush(@Suppress("UNUSED_PARAMETER") memoryCachesOnly: Boolean) {
        flush()
    }

    /**
     * DEPRECATED: This API should be removed because:
     *   - It's not obvious what "clean" means: It does not exactly describe the current implementation, and it also sounds similar to
     *   "clear" which means removing all the map entries, but this method does not do that.
     *   - This method currently calls [close] (and [deleteStorageFiles]). However, [close] is often already called separately and
     *   automatically, so this API makes it more likely for [close] to be accidentally called twice.
     *
     * Instead, just call [close] and/or [deleteStorageFiles] explicitly.
     */
    fun clean() {
        close()
        deleteStorageFiles()
    }

    @Synchronized
    private fun forEachMapSafe(actionName: String, action: (BasicMap<*, *>) -> Unit) {
        val actionExceptions = LinkedHashMap<String, Exception>()
        maps.forEach {
            try {
                action(it)
            } catch (e: Exception) {
                actionExceptions[it.storageFile.name] = e
            }
        }
        if (actionExceptions.isNotEmpty()) {
            val desc = "Could not $actionName incremental caches in $cachesDir: ${actionExceptions.keys.joinToString(", ")}"
            val allIOExceptions = actionExceptions.all { it is IOException }
            val ex = if (allIOExceptions) IOException(desc) else Exception(desc)
            actionExceptions.forEach { (_, e) -> ex.addSuppressed(e) }
            throw ex
        }
    }

    @TestOnly
    @Synchronized
    fun dump(): String = maps.joinToString("\n\n") { it.dump() }
}